import json

from ..HTTP.Requests import *

BASE_URL = 'https://www.googleapis.com/gmail/v1/'
OAUTH_BASE_URL = 'https://oauth2.googleapis.com/'

class gmail:

  def __init__(self, client_id, client_secret, refresh_token):
    self.refresh_token = refresh_token
    self.headers = {'Content-Type': 'application/json', 'Accept': 'application/json'}
    auth_body = {
        'grant_type': 'refresh_token',
        'client_id': client_id,
        'client_secret': client_secret,
        'refresh_token': refresh_token
    }
    auth_url = f"{OAUTH_BASE_URL}token"
    result = post(auth_url, self.headers, json.dumps(auth_body))
    self.access_token = result['content'].get('access_token')
    self.headers['Authorization'] = f"Bearer {self.access_token}"

  def get_auth_headers(self):
    return self.headers

  def get_me(self):
    post_url = f"{BASE_URL}users/me/profile"
    result = get(post_url, self.headers)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to retrieve profile. Result: {result}")
    return result

  def watch_me(self, label_ids=None):
    post_url = f"{BASE_URL}users/me/watch"
    post_body = {
        'topicName': 'projects/expressintegrations/topics/gmail',
        'labelIds': label_ids if label_ids else ['INBOX']
    }
    result = post(post_url, self.headers, json.dumps(post_body))
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to watch inbox. Result: {result}")
    return result

  def stop_watching_me(self):
    post_url = f"{BASE_URL}users/me/stop"
    result = post(post_url, self.headers)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to stop watching inbox. Result: {result}")
    return result

  def revoke_token(self):
    headers = {
        'Content-type': 'application/x-www-form-urlencoded'
    }
    post_url = f"{OAUTH_BASE_URL}revoke?token={self.refresh_token}"
    result = post(post_url, headers)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to revoke token. Result: {result}")
    return result

  def get_history_list(self, start_history_id):
    post_url = f"{BASE_URL}users/me/history?startHistoryId={start_history_id}"
    result = get(post_url, self.headers)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to retrieve history list. Result: {result}")
    return result

  def get_labels(self):
    post_url = f"{BASE_URL}users/me/labels"
    result = get(post_url, self.headers)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to retrieve labels. Result: {result}")
    return result

  def get_message(self, message_id):
    post_url = f"{BASE_URL}users/me/messages/{message_id}?format=full"
    result = get(post_url, self.headers)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to retrieve message. Result: {result}")
    return result

  def modify_message(self, message_id, add_label_ids=None, remove_label_ids=None):
    if not add_label_ids and not remove_label_ids:
      raise Exception(f"You must provide either a list of label IDs to add or remove.")
    post_url = f"{BASE_URL}users/me/messages/{message_id}/modify"
    post_body = {}
    if add_label_ids:
      post_body['addLabelIds'] = add_label_ids

    if remove_label_ids:
      post_body['removeLabelIds'] = remove_label_ids
    result = post(post_url, self.headers, json.dumps(post_body))
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to modify message. Result: {result}")
    return result



import requests
import json
import time
from ..Utils import Utils

URL = 'url'
HEADERS = 'headers'
PAYLOAD = 'payload'
STATUS_CODE = 'status_code'
CONTENT = 'content'


def response_is_retryable(result):
  return Utils.is_retryable(result[STATUS_CODE])


def post(url, headers, payload=None, auth=None):
  result = requests.post(url, data=payload, headers=headers, auth=auth)
  wait = 1
  while Utils.is_retryable(result.status_code):
    time.sleep(wait)
    wait = wait * 2 if wait < 10 else 10
    result = requests.post(url, data=payload, headers=headers, auth=auth)
  content = None
  if result.text:
    content = json.loads(result.text)
  return {
      URL: url,
      HEADERS: headers,
      PAYLOAD: payload,
      STATUS_CODE: result.status_code,
      CONTENT: content
  }


def patch(url, headers, payload=None, auth=None):
  result = requests.patch(url, data=payload, headers=headers, auth=auth)
  wait = 1
  while Utils.is_retryable(result.status_code):
    time.sleep(wait)
    wait = wait * 2 if wait < 10 else 10
    result = requests.patch(url, data=payload, headers=headers, auth=auth)
  content = None
  if result.text:
    content = json.loads(result.text)
  return {
      URL: url,
      HEADERS: headers,
      PAYLOAD: payload,
      STATUS_CODE: result.status_code,
      CONTENT: content
  }


def get(url, headers, auth=None, params=None):
  result = requests.get(url, headers=headers, auth=auth, params=params)
  wait = 1
  while Utils.is_retryable(result.status_code):
    print(f"retrying. last result: {result}")
    time.sleep(wait)
    wait = wait * 2 if wait < 10 else 10
    result = requests.get(url, headers=headers, auth=auth)
  content = None
  if result.text:
    content = json.loads(result.text)
  return {
      URL: url,
      HEADERS: headers,
      STATUS_CODE: result.status_code,
      CONTENT: content
  }


def put(url, headers, payload=None, auth=None):
  result = requests.put(url, data=payload, headers=headers, auth=auth)
  wait = 1
  while Utils.is_retryable(result.status_code):
    time.sleep(wait)
    wait = wait * 2 if wait < 10 else 10
    result = requests.put(url, data=payload, headers=headers, auth=auth)
  content = None
  if result.text:
    content = json.loads(result.text)
  return {
      URL: url,
      HEADERS: headers,
      PAYLOAD: payload,
      STATUS_CODE: result.status_code,
      CONTENT: content
  }


def delete(url, headers, auth=None):
  result = requests.delete(url, headers=headers, auth=auth)
  wait = 1
  while Utils.is_retryable(result.status_code):
    time.sleep(wait)
    wait = wait * 2 if wait < 10 else 10
    result = requests.delete(url, headers=headers, auth=auth)
  content = None
  if result.text:
    content = json.loads(result.text)
  return {
      URL: url,
      HEADERS: headers,
      STATUS_CODE: result.status_code,
      CONTENT: content
  }



from ..HTTP.Requests import *
from .zoho import HEADERS
from .zoho import authenticate

BASE_URL = 'https://books.zoho.com/api/v3/expenses'


def create_expense(expense):
  authenticate()
  post_url = f"{BASE_URL}"
  result = post(post_url, HEADERS, json.dumps(expense))
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = post(post_url, HEADERS, json.dumps(expense))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to create expense. Result: {result}")
  return result


def update_expense(expense_id, expense):
  authenticate()
  post_url = f"{BASE_URL}/{expense_id}"
  result = put(post_url, HEADERS, json.dumps(expense))
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = put(post_url, HEADERS, json.dumps(expense))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to update expense. Result: {result}")
  return result


def get_expense(expense_id):
  authenticate()
  post_url = f"{BASE_URL}/{expense_id}"
  result = get(post_url, HEADERS)
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to retrieve expense. Result: {result}")
  return result


def search_expenses(description=None,
                    reference_number=None,
                    date=None,
                    status=None,
                    amount=None,
                    account_name=None,
                    customer_name=None,
                    vendor_name=None,
                    customer_id=None,
                    vendor_id=None,
                    recurring_expense_id=None,
                    paid_through_account_id=None,
                    search_text=None,
                    filter_by=None,
                    sort_column='created_time',
                    **kwargs):
  authenticate()
  post_url = f"{BASE_URL}?sort_column={sort_column}"
  if description is not None:
    post_url = f"{post_url}&description={description}"
  if reference_number is not None:
    post_url = f"{post_url}&reference_number={reference_number}"
  if date is not None:
    post_url = f"{post_url}&date={date}"
  if status is not None:
    post_url = f"{post_url}&status={status}"
  if amount is not None:
    post_url = f"{post_url}&amount={amount}"
  if account_name is not None:
    post_url = f"{post_url}&account_name={account_name}"
  if customer_name is not None:
    post_url = f"{post_url}&customer_name={customer_name}"
  if vendor_name is not None:
    post_url = f"{post_url}&vendor_name={vendor_name}"
  if customer_id is not None:
    post_url = f"{post_url}&customer_id={customer_id}"
  if vendor_id is not None:
    post_url = f"{post_url}&vendor_id={vendor_id}"
  if recurring_expense_id is not None:
    post_url = f"{post_url}&recurring_expense_id={recurring_expense_id}"
  if paid_through_account_id is not None:
    post_url = f"{post_url}&paid_through_account_id={paid_through_account_id}"
  if filter_by is not None:
    post_url = f"{post_url}&filter_by={filter_by}"
  if search_text is not None:
    post_url = f"{post_url}&search_text={search_text}"
  for name, value in kwargs.items():
    post_url = f"{post_url}&{name}={value}"
  result = get(post_url, HEADERS)
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to retrieve expenses. Result: {result}")
  return result



from ..HTTP.Requests import *
from .zoho import HEADERS
from .zoho import authenticate

BASE_URL = 'https://books.zoho.com/api/v3/items'


def create_item(item):
  authenticate()
  post_url = f"{BASE_URL}"
  result = post(post_url, HEADERS, json.dumps(item))
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = post(post_url, HEADERS, json.dumps(item))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to create item. Result: {result}")
  return result


def update_item(item_id, item):
  authenticate()
  post_url = f"{BASE_URL}/{item_id}"
  result = put(post_url, HEADERS, json.dumps(item))
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = put(post_url, HEADERS, json.dumps(item))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to update item. Result: {result}")
  return result


def get_item(item_id):
  authenticate()
  post_url = f"{BASE_URL}/{item_id}"
  result = get(post_url, HEADERS)
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to retrieve item. Result: {result}")
  return result


def search_items(name=None,
                 description=None,
                 rate=None,
                 tax_id=None,
                 filter_by=None,
                 search_text=None,
                 sort_column='created_time',
                 **kwargs):
  authenticate()
  # filter_by Allowed Values: Status.All, Status.Active and Status.Inactive
  # sort_column Allowed Values: name, rate and tax_name
  post_url = f"{BASE_URL}?sort_column={sort_column}"
  if name is not None:
    post_url = f"{post_url}&name={name}"
  if description is not None:
    post_url = f"{post_url}&description={description}"
  if rate is not None:
    post_url = f"{post_url}&rate={rate}"
  if tax_id is not None:
    post_url = f"{post_url}&tax_id={tax_id}"
  if filter_by is not None:
    post_url = f"{post_url}&filter_by={filter_by}"
  if search_text is not None:
    post_url = f"{post_url}&search_text={search_text}"
  for name, value in kwargs.items():
    post_url = f"{post_url}&{name}={value}"
  result = get(post_url, HEADERS)
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to retrieve items. Result: {result}")
  return result



from ..HTTP.Requests import *
from .zoho import HEADERS
from .zoho import authenticate
from ..Utils import Utils

BASE_URL = 'https://books.zoho.com/api/v3/contacts'


def create_contact(contact):
  authenticate()
  result = post(BASE_URL, HEADERS, json.dumps(contact))
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = post(BASE_URL, HEADERS, json.dumps(contact))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to create contact. Result: {result}")
  return result


def update_contact(contact_id, contact):
  authenticate()
  post_url = f"{BASE_URL}/{contact_id}"
  result = put(post_url, HEADERS, json.dumps(contact))
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = put(post_url, HEADERS, json.dumps(contact))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to update contact. Result: {result}")
  return result


def get_contact(contact_id):
  authenticate()
  post_url = f"{BASE_URL}/{contact_id}"
  result = get(post_url, HEADERS)
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to retrieve contact. Result: {result}")
  return result


def search_contacts(contact_name=None,
                    company_name=None,
                    first_name=None,
                    last_name=None,
                    address=None,
                    email=None,
                    phone=None,
                    filter_by=None,
                    search_text=None,
                    sort_column='created_time',
                    **kwargs):
  authenticate()
  # filter_by Allowed Values: Status.All, Status.Active, Status.Inactive, Status.Duplicate and Status.Crm
  # sort_column Allowed Values: contact_name, first_name, last_name, email, outstanding_receivable_amount, created_time and last_modified_time
  post_url = f"{BASE_URL}?sort_column={sort_column}"
  if contact_name is not None:
    post_url = f"{post_url}&contact_name={contact_name}"
  if company_name is not None:
    post_url = f"{post_url}&company_name={company_name}"
  if first_name is not None:
    post_url = f"{post_url}&first_name={first_name}"
  if last_name is not None:
    post_url = f"{post_url}&last_name={last_name}"
  if address is not None:
    post_url = f"{post_url}&address={address}"
  if email is not None:
    post_url = f"{post_url}&email={email}"
  if phone is not None:
    post_url = f"{post_url}&phone={phone}"
  if filter_by is not None:
    post_url = f"{post_url}&filter_by={filter_by}"
  if search_text is not None:
    post_url = f"{post_url}&search_text={search_text}"
  for name, value in kwargs.items():
    post_url = f"{post_url}&{name}={value}"
  result = get(post_url, HEADERS)
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to retrieve contacts. Result: {result}")
  return result



from ..HTTP.Requests import *
from .hubspot import BASE_URL
from .hubspot import HEADERS


def get_properties(object_type):
  post_url = f"{BASE_URL}crm/v3/properties/{object_type}?archived=false"
  result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to get properties. Result: {result}")
  return result


def get_property(object_type, property_name):
  post_url = f"{BASE_URL}crm/v3/properties/{object_type}/{property_name}?archived=false"
  result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to get property. Result: {result}")
  return result


def update_property(object_type, property_name, data):
  post_url = f"{BASE_URL}crm/v3/properties/{object_type}/{property_name}"
  acceptable_keys = [
      'groupName',
      'hidden',
      'displayOrder',
      'options',
      'label',
      'type',
      'fieldType',
      'formField'
  ]
  data = {k: v for k, v in data.items() if k in acceptable_keys}
  result = patch(post_url, HEADERS, payload=json.dumps(data))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to update property. Result: {result}")
  return result



from ..HTTP.Requests import *
from .hubspot import BASE_URL
from .hubspot import HEADERS
from . import object_properties


def search_products_by_property_value(property_name, property_value, property_names=None, after=None):
  filters = [
      {
          'propertyName': property_name,
          'operator': 'EQ',
          'value': property_value
      }
  ]
  return search_products(property_names, filters, after)


def search_products_by_property_known(property_name, property_names=None, after=None):
  filters = [
      {
          'propertyName': property_name,
          'operator': 'HAS_PROPERTY'
      }
  ]
  return search_products(property_names, filters, after)


def search_products_by_property_less_than(property_name, property_value, property_names=None, after=None):
  filters = [
      {
          'propertyName': property_name,
          'operator': 'LT',
          'value': property_value
      }
  ]
  return search_products(property_names, filters, after)


def search_products_by_property_greater_than(property_name, property_value, property_names=None, after=None):
  filters = [
      {
          'propertyName': property_name,
          'operator': 'GT',
          'value': property_value
      }
  ]
  return search_products(property_names, filters, after)


def search_products_by_property_values(property_name, property_values, property_names=None, after=None):
  filters = [
      {
          'propertyName': property_name,
          'operator': 'IN',
          'values': property_values
      }
  ]
  return search_products(property_names, filters, after)


def search_products(property_names, filters, after=None):
  post_url = f"{BASE_URL}crm/v3/objects/products/search"
  if after:
    post_url = f"{post_url}?after={after}"
  post_body = {
      'filterGroups': [
          {
              'filters': filters
          }
      ],
      'sorts': [],
      'limit': 100
  }
  if property_names:
    post_body['properties'] = property_names
  result = post(post_url, HEADERS, json.dumps(post_body))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to search products. Result: {result}")
  return result


def get_product(product_id, property_names=None, associations=None):
  post_url = f"{BASE_URL}crm/v3/objects/products/{product_id}?archived=false"
  if property_names:
    post_url = f"{post_url}&properties={'%2C'.join(property_names)}"

  if associations:
    post_url = f"{post_url}&associations={'%2C'.join(associations)}"
  result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to retrieve product. Result: {result}")
  return result


def create_product(properties):
  post_url = f"{BASE_URL}crm/v3/objects/products"
  post_body = {
      'properties': properties
  }
  result = post(post_url, HEADERS, json.dumps(post_body))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to create product. Result: {result}")
  return result


def update_product(product_id, properties):
  post_url = f"{BASE_URL}crm/v3/objects/products/{product_id}?"
  post_body = {
      'properties': properties
  }
  result = patch(post_url, HEADERS, json.dumps(post_body))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to update product. Result: {result}")
  return result


def get_product_properties():
  return object_properties.get_properties('products')


def get_product_property(property_name):
  return object_properties.get_property('products', property_name)



from ..HTTP.Requests import *
from .hubspot import BASE_URL
from .hubspot import HEADERS


def get_associations(from_object_type, to_object_type, from_object_id):
  post_url = f"{BASE_URL}crm/v3/associations/{from_object_type}/{to_object_type}/batch/read"
  post_body = {
      'inputs': [
          {
              'id': from_object_id
          }
      ]
  }
  result = post(post_url, HEADERS, json.dumps(post_body))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to retrieve associations. Result: {result}")
  return result


def set_parent_company(company_id, parent_company_id):
  post_url = f"{BASE_URL}crm/v3/objects/companies/{company_id}/associations/company/{parent_company_id}/CHILD_TO_PARENT_COMPANY"
  result = put(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to set parent company. Result: {result}")
  return result


def set_child_company(company_id, child_company_id):
  post_url = f"{BASE_URL}crm/v3/objects/companies/{company_id}/associations/company/{child_company_id}/PARENT_TO_CHILD_COMPANY"
  result = put(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to set child company. Result: {result}")
  return result


def set_company_for_contact(contact_id, company_id):
  post_url = f"{BASE_URL}crm/v3/objects/contacts/{contact_id}/associations/company/{company_id}/CONTACT_TO_COMPANY"
  result = put(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to set company for contact. Result: {result}")
  return result


def set_company_for_deal(deal_id, company_id):
  post_url = f"{BASE_URL}crm/v3/objects/deals/{deal_id}/associations/company/{company_id}/DEAL_TO_COMPANY"
  result = put(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to set company for deal. Result: {result}")
  return result


def delete_company_from_deal(deal_id, company_id):
  post_url = f"{BASE_URL}crm/v3/objects/deals/{deal_id}/associations/company/{company_id}/DEAL_TO_COMPANY"
  result = delete(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to delete company from deal. Result: {result}")
  return result


def set_contact_for_deal(deal_id, contact_id):
  post_url = f"{BASE_URL}crm/v3/objects/deals/{deal_id}/associations/contact/{contact_id}/DEAL_TO_CONTACT"
  result = put(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to set contact for deal. Result: {result}")
  return result


def set_deal_for_line_item(line_item_id, deal_id):
  post_url = f"{BASE_URL}crm/v3/objects/line_items/{line_item_id}/associations/deals/{deal_id}/LINE_ITEM_TO_DEAL"
  result = put(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to set deal for line item. Result: {result}")
  return result



from ..HTTP.Requests import *
from ..Utils import Utils

HEADERS = {'Content-Type': 'application/json', 'Accept': 'application/json'}
AUTH_HEADERS = {'Content-Type': 'application/x-www-form-urlencoded'}
AUTH_URL = 'https://accounts.zoho.com'
BASE_URL = 'https://books.zoho.com/api/v3/{}?organization_id='
CONFIG = {}


def authenticate(force=False):
  if CONFIG['expires_at'] is None or CONFIG['expires_at'] < Utils.now() or force:
    post_url = f"{AUTH_URL}/oauth/v2/token"
    auth_body = {
        'grant_type': 'refresh_token',
        'client_id': CONFIG['client_id'],
        'client_secret': CONFIG['client_secret'],
        'refresh_token': CONFIG['refresh_token']
    }
    result = post(post_url, AUTH_HEADERS, auth_body)
    auth_result = result['content']
    if 'access_token' not in auth_result.keys():
      raise Exception(f"Access token not returned: {result}")
    CONFIG['access_token'] = auth_result['access_token']
    HEADERS['Authorization'] = f"Zoho-oauthtoken {CONFIG['access_token']}"
    CONFIG['expires_at'] = Utils.now() + (auth_result['expires_in'] * 1000)


class zoho:
  import json
  from . import invoices
  from . import contacts
  from . import customerpayments
  from . import users
  from . import items
  from . import expenses

  def __init__(self, auth_data):
    CONFIG['refresh_token'] = auth_data.get('refresh_token')
    CONFIG['access_token'] = auth_data.get('access_token')
    CONFIG['expires_at'] = auth_data.get('expires_at')
    CONFIG['client_id'] = auth_data.get('client_id')
    CONFIG['client_secret'] = auth_data.get('client_secret')
    authenticate()

  # Contains utilities for interacting with the Zoho API

  def get_org_details(self):
    authenticate()
    post_url = f"https://www.zohoapis.com/crm/v2/org"
    HEADERS['Authorization'] = f"Zoho-oauthtoken {self.get_auth_data()['access_token']}"
    return get(post_url, HEADERS)

  def revoke_token(self):
    post_url = f"{AUTH_URL}/oauth/v2/token/revoke?token={CONFIG['refresh_token']}"
    return post(post_url, AUTH_HEADERS)

  def get_auth_data(self):
    return {
        'access_token': CONFIG['access_token'],
        'expires_at': CONFIG['expires_at']
    }



import time
from datetime import datetime, timezone
import calendar
import smtplib
from email.message import EmailMessage

from collections import defaultdict

epoch = datetime.utcfromtimestamp(0).replace(tzinfo=timezone.utc)


def access_secret_version(project_id, secret_id, version_id):
  """
  Access the payload for the given secret version if one exists. The version
  can be a version number as a string (e.g. "5") or an alias (e.g. "latest").
  """

  # Import the Secret Manager client library.
  from google.cloud import secretmanager

  # Create the Secret Manager client.
  client = secretmanager.SecretManagerServiceClient()

  # Build the resource name of the secret version.
  secret_version_path = f'projects/{project_id}/secrets/{secret_id}/versions/{version_id}'
  #name = client.secret_version_path(project_id, secret_id, version_id)

  # Access the secret version.
  response = client.access_secret_version(request={'name': secret_version_path})
  payload = response.payload.data.decode('UTF-8')
  return payload


def send_email(sender, pw, recipient, subject, message):
  msg = EmailMessage()
  msg.set_content(message)
  msg['Subject'] = subject
  msg['From'] = sender
  msg['To'] = recipient

  server = smtplib.SMTP_SSL('smtp.gmail.com', 465)
  server.login(sender, pw)
  server.send_message(msg)
  server.quit()


def format_date_time(epoch_ms):
  if not epoch_ms:
    return None
  return datetime.utcfromtimestamp(int(epoch_ms) // 1000).replace(microsecond=int(epoch_ms) % 1000 * 1000)


def date_from_epoch_ms(epoch_ms):
  date_time = format_date_time(epoch_ms)
  if not date_time:
    return None
  return date_time.date()


def closest_date_next_month(year, month, day):
  month = month + 1
  if month == 13:
    month = 1
    year = year + 1

  condition = True
  while condition:
    try:
      return datetime(year, month, day)
    except ValueError:
      day = day - 1
    condition = day > 26

  raise Exception('Problem getting date next month')


def pass_job(job, start_time, success, event, error, results, job_history=None):
  if job_history:
    job_history.update(
        start_time=start_time,
        end_time=datetime.now(),
        success=success,
        input=event,
        error=error,
        results=results
    )
  else:
    job_history = app_tables.job_history.add_row(
        job=job,
        start_time=start_time,
        end_time=datetime.now(),
        success=success,
        input=event,
        error=error,
        results=results
    )
  return job_history


def fail_job(job, job_name, user_email, start_time, event, error, results, stack_trace, job_history=None):
  if job_history:
    job_history.update(
        start_time=start_time,
        end_time=datetime.now(),
        success=False,
        input=event,
        error=error,
        results=results
    )
  else:
    job_history = app_tables.job_history.add_row(
        job=job,
        start_time=start_time,
        end_time=datetime.now(),
        success=False,
        input=event,
        error=error,
        results=results
    )
  anvil.server.call('send_admin_error_email', job_name, user_email, start_time, event, error, stack_trace)
  return job_history


def start_admin_job(job_history):
  job_start_time = now()
  job_history.update(last_job_start_time=format_date_time(job_start_time), status="running...")
  return job_start_time


def pass_admin_job(job_history, job_start_time, results):
  job_end_time = now()
  job_history.update(
      last_job_run_results=results,
      last_job_end_time=format_date_time(job_end_time),
      status="completed successfully",
      last_successful_run_time=format_date_time(job_end_time),
      last_job_duration=round((job_end_time - job_start_time) / 1000, 3)
  )


def fail_admin_job(job_history, job_start_time, results, message):
  job_end_time = now()
  job_history.update(
      last_job_run_results=results,
      last_job_end_time=format_date_time(job_end_time),
      status=f"failed: {message}",
      last_job_duration=round((job_end_time - job_start_time) / 1000, 3)
  )


def now():
  return round(time.time() * 1000)


def getEpochMillis(dt):
  return (dt - epoch).total_seconds() * 1000


def validateDateAndReturnEpochTimeInMillis(date):
  date = str(date).replace("-", "/").replace(".", "/")
  yyyy, mm, dd = date.split('/')
  dd = int(dd.lstrip("0"))
  mm = int(mm.lstrip("0"))
  yyyy = int(yyyy)
  if mm in [1, 3, 5, 7, 8, 10, 12]:
    maxDay = 31
  elif mm in [4, 6, 9, 11]:
    maxDay = 30
  elif yyyy % 4 == 0 and yyyy % 100 != 0 or yyyy % 400 == 0:
    maxDay = 29
  else:
    maxDay = 28

  if mm < 1 or mm > 12:
    raise Exception("Please specify a valid month.")
  elif dd < 1 or dd > maxDay:
    raise Exception("Please specify a valid day.")
  elif len(f"{yyyy}") != 4:
    raise Exception("Please specify a valid 4-digit year.")
  else:
    dateTime = f"{date} 00:00:00"
    epoch = int(time.mktime(time.strptime(dateTime, "%Y/%m/%d %H:%M:%S"))) * 1000

  return epoch


def partition(seq, key):
  d = {}
  for x in seq:
    d[key(x)] = []
  return d


def makehash():
  return defaultdict(makehash)


def get_filtered_dict(search_term, dictionary):
  return {k: v for k, v in dictionary.items() if any(search_term in str(x).lower() for x in v.values())}


def is_success(status_code):
  return str(status_code).startswith('2')


def is_not_found(status_code):
  return str(status_code) == '404'


def is_retryable(status_code):
  return str(status_code) in ['502', '429'] or not status_code


def is_unauthorized(status_code):
  return str(status_code) in ['401', '403']


def validate_response(response, error_message, success_message=None):
  if not is_success(response['status_code']):
    response['error_message'] = '%s %s:%s' % (error_message, response['status_code'], response['content'])
  elif success_message != None:
    response['success_message'] = success_message
  return response


def remove_keys_from_dict(d, keys):
  return {k: v for k, v in d.items() if k not in keys}


def remove_keys_from_dict_except(d, keys):
  return {k: v for k, v in d.items() if k in keys}


def apply_func_to_items(iterable, func):
  return map(lambda x: func(x), iterable)



import json

from ..HTTP.Requests import *
from ..Utils import Utils
from urllib.parse import urlencode


class ministry:

  def __init__(self, site, access_token):
    self.headers = {"Content-Type": "application/json", "Accept": "application/json"}
    self.headers['Authorization'] = f"Bearer {access_token}"
    self.base_url = site

  def create_record(self, table, record):
    return self.create_records(table, [record])

  def create_records(self, table, records):
    post_url = f"{self.base_url}tables/{table}"
    result = post(post_url, self.headers, json.dumps(records))
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to create record. Result: {result}")
    return result

  def get_record_by_id(self, table, record_id):
    post_url = f"{self.base_url}tables/{table}/{record_id}"
    result = get(post_url, self.headers)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to retrieve record. Result: {result}")
    return result

  def search_records(self, table, search_string=None, order_by=None):
    post_url = f"{self.base_url}tables/{table}"
    if search_string is not None:
      sql_filter = {'filter': search_string}
      post_url = f"{post_url}?${urlencode(sql_filter)}"

    if order_by is not None:
      sql_order = {'order_by': order_by}
      post_url = f"{post_url}&${urlencode(sql_order)}"
    result = get(post_url, self.headers)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to retrieve records. Result: {result}")
    return result

  def update_record(self, table, record):
    return self.update_records(table, [record])

  def update_records(self, table, records):
    post_url = f"{self.base_url}tables/{table}"
    result = put(post_url, self.headers, json.dumps(records))
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to update record. Result: {result}")
    return result

  def delete_record(self, table, record_id):
    post_url = f"{self.base_url}tables/{table}/{record_id}"
    result = delete(post_url, self.headers)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to delete record. Result: {result}")
    return result



from ..HTTP.Requests import *

HEADERS = {"Content-Type": "application/json", "Accept": "application/json"}
AUTH_HEADERS = {"Content-Type": "application/x-www-form-urlencoded"}
BASE_URL = "https://api.hubapi.com/"


class hubspot:
  import json
  from . import associations
  from . import companies
  from . import contacts
  from . import deals
  from . import owners
  from . import products
  from . import lineitems

  def __init__(self, client_id, client_secret, refresh_token):
    self.refresh_token = refresh_token
    post_url = f"{BASE_URL}oauth/v1/token"
    auth_body = {
        'grant_type': 'refresh_token',
        'client_id': client_id,
        'client_secret': client_secret,
        'refresh_token': refresh_token
    }
    result = post(post_url, AUTH_HEADERS, auth_body)
    auth_result = result['content']
    HEADERS['Authorization'] = f"Bearer {auth_result['access_token']}"

  # Contains utilities for interacting with the HubSpot API
  def get_token_details(self):
    post_url = f"{BASE_URL}oauth/v1/refresh-tokens/{self.refresh_token}"
    return get(post_url, HEADERS)

  def revoke_token(self):
    post_url = f"{BASE_URL}oauth/v1/refresh-tokens/{self.refresh_token}"
    return delete(post_url, AUTH_HEADERS)



from ..HTTP.Requests import *
from .zoho import HEADERS
from .zoho import authenticate

BASE_URL = 'https://books.zoho.com/api/v3/invoices'


def create_invoice(invoice):
  authenticate()
  post_url = f"{BASE_URL}"
  result = post(post_url, HEADERS, json.dumps(invoice))
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = post(post_url, HEADERS, json.dumps(invoice))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to create invoice. Result: {result}")
  return result


def update_invoice(invoice_id, invoice):
  authenticate()
  post_url = f"{BASE_URL}/{invoice_id}"
  result = put(post_url, HEADERS, json.dumps(invoice))
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = put(post_url, HEADERS, json.dumps(invoice))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to update invoice. Result: {result}")
  return result


def get_invoice(invoice_id):
  authenticate()
  post_url = f"{BASE_URL}/{invoice_id}"
  result = get(post_url, HEADERS)
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to retrieve invoice. Result: {result}")
  return result


def search_invoices(invoice_number=None,
                    item_name=None,
                    item_id=None,
                    item_description=None,
                    reference_number=None,
                    customer_name=None,
                    recurring_invoice_id=None,
                    email=None,
                    total=None,
                    balance=None,
                    date=None,
                    due_date=None,
                    status=None,
                    customer_id=None,
                    filter_by=None,
                    search_text=None,
                    sort_column='created_time',
                    **kwargs):
  authenticate()
  # filter_by Allowed Values: Status.All, Status.Sent, Status.Draft, Status.OverDue, Status.Paid, Status.Void, Status.Unpaid, Status.PartiallyPaid, Status.Viewed and Date.PaymentExpectedDate
  # sort_column Allowed Values: customer_name, invoice_number, date, due_date, total, balance and created_time
  post_url = f"{BASE_URL}?sort_column={sort_column}"
  if invoice_number is not None:
    post_url = f"{post_url}&invoice_number={invoice_number}"
  if item_name is not None:
    post_url = f"{post_url}&item_name={item_name}"
  if item_id is not None:
    post_url = f"{post_url}&item_id={item_id}"
  if item_description is not None:
    post_url = f"{post_url}&item_description={item_description}"
  if reference_number is not None:
    post_url = f"{post_url}&reference_number={reference_number}"
  if customer_name is not None:
    post_url = f"{post_url}&customer_name={customer_name}"
  if recurring_invoice_id is not None:
    post_url = f"{post_url}&recurring_invoice_id={recurring_invoice_id}"
  if email is not None:
    post_url = f"{post_url}&email={email}"
  if total is not None:
    post_url = f"{post_url}&total={total}"
  if balance is not None:
    post_url = f"{post_url}&balance={balance}"
  if date is not None:
    post_url = f"{post_url}&date={date}"
  if due_date is not None:
    post_url = f"{post_url}&due_date={due_date}"
  if status is not None:
    post_url = f"{post_url}&status={status}"
  if customer_id is not None:
    post_url = f"{post_url}&customer_id={customer_id}"
  if filter_by is not None:
    post_url = f"{post_url}&filter_by={filter_by}"
  if search_text is not None:
    post_url = f"{post_url}&search_text={search_text}"
  for name, value in kwargs.items():
    post_url = f"{post_url}&{name}={value}"
  result = get(post_url, HEADERS)
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to retrieve invoices. Result: {result}")
  return result



from ..HTTP.Requests import *
from .hubspot import BASE_URL
from .hubspot import HEADERS


def get_owner(owner_id):
  post_url = f"{BASE_URL}crm/v3/owners/{owner_id}"
  result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to retrieve owner. Result: {result}")
  return result


def search_owners(email=None, limit=100, after=None):
  post_url = f"{BASE_URL}crm/v3/owners?limit={limit}"
  if email is not None:
    post_url = f"{post_url}&email={email}"
  if after is not None:
    post_url = f"{post_url}&after={after}"
  result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to search owners. Result: {result}")
  return result



from ..HTTP.Requests import *


class managed_missions:
  def __init__(self, api_key):
    self.headers = {"Content-Type": "application/json", "Accept": "application/json"}
    self.headers['Authorization'] = f"Bearer {api_key}"
    self.base_url = 'https://app.managedmissions.com/API'

  # Contains utilities for interacting with the Managed Missions api
  def get_active_mission_trips(self, **kw):
    post_url = f"{self.base_url}/MissionTripAPI/ActiveTrips"
    result = get(post_url, self.headers, params=kw)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to retrieve active mission trips. Result: {result}")
    return result

  def get_mission_trips(self, **kw):
    post_url = f"{self.base_url}/MissionTripAPI/List"
    result = get(post_url, self.headers, params=kw)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to retrieve mission trips. Result: {result}")
    return result

  def get_trip_members(self, mission_trip_id, **kw):
    post_url = f"{self.base_url}/MemberAPI/GetMembersOfTrip"
    kw['MissionTripId'] = mission_trip_id
    result = get(post_url, self.headers, params=kw)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to retrieve trip members. Result: {result}")
    return result

  def get_people(self, **kw):
    post_url = f"{self.base_url}/PersonAPI/List"
    result = get(post_url, self.headers, params=kw)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to retrieve people. Result: {result}")
    return result

  def get_person(self, **kw):
    post_url = f"{self.base_url}/PersonAPI/Get"
    result = get(post_url, self.headers, params=kw)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to retrieve person. Result: {result}")
    return result

  def get_contributions(self, **kw):
    post_url = f"{self.base_url}/ContributionAPI/List"
    result = get(post_url, self.headers, params=kw)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to retrieve contributions. Result: {result}")
    return result

  def get_donors(self, **kw):
    post_url = f"{self.base_url}/DonorAPI/List"
    result = get(post_url, self.headers, params=kw)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to retrieve donors. Result: {result}")
    return result

  def get_trip_donors(self, mission_trip_id, **kw):
    post_url = f"{self.base_url}/DonorAPI/GetByMissionTrip"
    kw['missionTripId'] = mission_trip_id
    result = get(post_url, self.headers, params=kw)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to retrieve trip donors. Result: {result}")
    return result



from ..HTTP.Requests import *
from .zoho import HEADERS
from .zoho import authenticate

BASE_URL = 'https://books.zoho.com/api/v3'


def create_user(user):
  authenticate()
  post_url = f"{BASE_URL}/users"
  result = post(post_url, HEADERS, json.dumps(user))
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = post(post_url, HEADERS, json.dumps(user))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to create user. Result: {result}")
  return result


def update_user(user_id, user):
  authenticate()
  post_url = f"{BASE_URL}/users/{user_id}"
  result = put(post_url, HEADERS, json.dumps(user))
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = put(post_url, HEADERS, json.dumps(user))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to update user. Result: {result}")
  return result


def get_user(user_id):
  authenticate()
  post_url = f"{BASE_URL}/users/{user_id}"
  result = get(post_url, HEADERS)
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to retrieve user. Result: {result}")
  return result


def search_users(filter_by=None, sort_column='email', **kwargs):
  authenticate()
  # sort_column Allowed Values: name, email, user_role and status
  post_url = f"{BASE_URL}/users?sort_column={sort_column}"
  if filter_by is not None:
    post_url = f"{post_url}&filter_by={filter_by}"
  for name, value in kwargs.items():
    post_url = f"{post_url}&{name}={value}"
  result = get(post_url, HEADERS)
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to retrieve users. Result: {result}")
  return result


def get_salesperson(salesperson_id):
  authenticate()
  SALESPERSON_ID = 'salesperson_id'
  CONTENT = 'content'
  DATA = 'data'
  result = search_salespersons()
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = search_salespersons()
  salespersons = {item[SALESPERSON_ID]: item for item in result[CONTENT][DATA]}
  return salespersons.get(salesperson_id)


def search_salespersons():
  authenticate()
  post_url = f"{BASE_URL}/salespersons"
  result = get(post_url, HEADERS)
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to search salespersons. Result: {result}")
  return result


def create_salesperson(name, email):
  authenticate()
  post_url = f"{BASE_URL}/salespersons"
  post_body = {
      'salesperson_name': name,
      'salesperson_email': email
  }
  result = post(post_url, HEADERS, json.dumps(post_body))
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = post(post_url, HEADERS, json.dumps(post_body))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to create salesperson. Result: {result}")
  return result



from ringcentral import SDK
import time


BASE_URL = 'https://platform.ringcentral.com'


class ringcentral:
  def __init__(self, client_id, client_secret, username, password, extension, server_url=BASE_URL, auth_data=None):
    rcsdk = SDK(client_id, client_secret, server_url)
    self.platform = rcsdk.platform()
    self.extension = extension
    if auth_data is not None:
      self.platform._auth.set_data(auth_data)

    if not self.platform.logged_in():
      self.platform.login(username, extension, password)

    try:
      self.get_account_details()
    except Exception as e:
      if 'Token not found' in str(e):
        self.platform.login(username, extension, password)
    self.account_id = self.get_account_details().json().id

  # Contains utilities for interacting with the RingCentral api
  def retry(self, method, url, failure_message, query_params=None, payload=None):
    result = None
    try:
      if method.lower() == 'get':
        result = self.platform.get(url, query_params)
      elif method.lower() == 'post':
        result = self.platform.post(url, payload, query_params)
      elif method.lower() == 'put':
        result = self.platform.put(url, payload, query_params)
      elif method.lower() == 'delete':
        result = self.platform.delete(url)
    except Exception as e:
      wait = 60
      error_message = str(e)
      while (
          'Request rate exceeded' in error_message
          or '429' in error_message
      ):
        print(f"retrying. last result: {str(e)}")
        time.sleep(wait)
        wait = wait * 2 if wait < 1000 else 1000
        try:
          if method.lower() == 'get':
            result = self.platform.get(url, query_params)
          elif method.lower() == 'post':
            result = self.platform.post(url, payload, query_params)
          elif method.lower() == 'put':
            result = self.platform.put(url, payload, query_params)
          elif method.lower() == 'delete':
            result = self.platform.delete(url)
        except Exception as e1:
          error_message = str(e1)
      if result is None or not result.ok:
        raise Exception(f"{failure_message} Result: {result.error() if result is not None else error_message}")
    return result

  def get_auth_data(self):
    return self.platform._auth.data()

  def get_account_details(self, account_id='~'):
    return self.retry(
        'GET',
        f"/restapi/v1.0/account/{account_id}",
        'Failed to get account details.'
    )

  def send_sms(self, sender_number, recipient_number, text, extension='~'):
    post_body = {
        'from': {
            'phoneNumber': sender_number
        },
        'to': [
            {
                'phoneNumber': recipient_number
            }
        ],
        'text': text
    }
    return self.retry(
        'POST',
        f"/restapi/v1.0/account/~/extension/{extension}/sms",
        'Failed to send sms.',
        payload=post_body
    )

  def renew_webhook(self, subscription_id):
    return self.retry(
        'POST',
        f"/restapi/v1.0/subscription/{subscription_id}/renew",
        'Failed to renew webhook.'
    )

  def create_contact(self, contact, dialing_plan=None):
    query_params = {}
    if dialing_plan is not None:
      query_params['dialingPlan'] = dialing_plan
    return self.retry(
        'POST',
        f"/restapi/v1.0/account/~/extension/~/address-book/contact",
        'Failed to create contact.',
        query_params=query_params,
        payload=contact
    )

  def get_contacts(self, starts_with=None, sort_by=None, page=None, per_page=None, phone_number=None):
    query_params = {}
    if starts_with is not None:
      query_params['startsWith'] = starts_with
    if sort_by is not None:
      query_params['sortBy'] = sort_by
    if page is not None:
      query_params['page'] = page
    if per_page is not None:
      query_params['perPage'] = per_page
    if phone_number is not None:
      query_params['phoneNumber'] = phone_number
    return self.retry(
        'GET',
        f"/restapi/v1.0/account/~/extension/~/address-book/contact",
        'Failed to search contacts.',
        query_params=query_params
    )

  def get_contact(self, contact_id):
    return self.retry(
        'GET',
        f"/restapi/v1.0/account/~/extension/~/address-book/contact/{contact_id}",
        'Failed to get contact.'
    )

  def update_contact(self, contact_id, contact, dialing_plan=None):
    query_params = {}
    if dialing_plan is not None:
      query_params['dialingPlan'] = dialing_plan
    return self.retry(
        'PUT',
        f"/restapi/v1.0/account/~/extension/~/address-book/contact/{contact_id}",
        'Failed to update contact.',
        query_params=query_params,
        payload=contact
    )

  def delete_contact(self, contact_id):
    return self.retry(
        'DELETE',
        f"/restapi/v1.0/account/~/extension/~/address-book/contact/{contact_id}",
        'Failed to delete contact.'
    )

  def search_company_directory(self, search_string=None, show_federated=True, extension_type=None, order_by=None, page=None, per_page=None):
    post_body = {}
    if search_string is not None:
      post_body['searchString'] = search_string
    post_body['showFederated'] = show_federated
    if extension_type is not None:
      post_body['extensionType'] = extension_type
    if order_by is not None:
      post_body['orderBy'] = order_by
    if page is not None:
      post_body['page'] = page
    if per_page is not None:
      post_body['perPage'] = per_page
    return self.retry(
        'POST',
        f"/restapi/v1.0/account/~/directory/entries/search",
        'Failed to search company directory.',
        payload=post_body
    )



import hashlib
import json

from ..HTTP.Requests import *


class mailchimp:
  def __init__(self, access_token, api_endpoint):
    self.headers = {"Content-Type": "application/json", "Accept": "application/json"}
    self.headers['Authorization'] = f"OAuth {access_token}"
    self.base_url = f"{api_endpoint}/3.0/"

  # Contains utilities for interacting with the MailChimp api
  def get_account_details(self):
    post_url = f"{self.base_url}"
    result = get(post_url, self.headers)
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to retrieve account details. Result: {result}")
    return result

  def get_list_member(self, list_id, subscriber_hash):
    post_url = f"{self.base_url}lists/{list_id}/members/{subscriber_hash}"
    result = get(post_url, self.headers)
    if not Utils.is_success(result['status_code']) and not Utils.is_not_found(result['status_code']):
      raise Exception(f"Failed to retrieve list member. Result: {result}")
    return result

  def get_list_member_by_email(self, list_id, email):
    subscriber_hash = hashlib.md5(email.encode()).hexdigest()
    post_url = f"{self.base_url}lists/{list_id}/members/{subscriber_hash}"
    result = get(post_url, self.headers)
    if not Utils.is_success(result['status_code']) and not Utils.is_not_found(result['status_code']):
      raise Exception(f"Failed to retrieve list member. Result: {result}")
    return result

  def create_list_member(self, list_id, email, status, merge_fields=None):
    post_url = f"{self.base_url}lists/{list_id}/members"
    post_body = {
        'email_address': email,
        'status': status,
        'merge_fields': merge_fields
    }
    result = post(post_url, self.headers, json.dumps(post_body))
    if not Utils.is_success(result['status_code']):
      raise Exception(f"Failed to create list member. Result: {result}")
    return result

  def update_list_member(self, list_id, subscriber_hash, email=None, status=None, merge_fields=None):
    post_url = f"{self.base_url}lists/{list_id}/members/{subscriber_hash}"
    post_body = {}
    if status is not None:
      post_body['status'] = status
    if merge_fields is not None:
      post_body['merge_fields'] = merge_fields
    if email:
      post_body['email_address'] = email
    result = put(post_url, self.headers, json.dumps(post_body))
    if not Utils.is_success(result['status_code']):
      if 'has signed up to a lot of lists very recently' in result['content']['detail']:
        return result
      else:
        raise Exception(f"Failed to update list member. Result: {result}")
    return result



from ..HTTP.Requests import *
from .zoho import HEADERS
from .zoho import authenticate

BASE_URL = 'https://books.zoho.com/api/v3/customerpayments'


def create_customerpayment(customerpayment):
  authenticate()
  post_url = f"{BASE_URL}"
  result = post(post_url, HEADERS, json.dumps(customerpayment))
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = post(post_url, HEADERS, json.dumps(customerpayment))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to create customerpayment. Result: {result}")
  return result


def update_customerpayment(customerpayment_id, customerpayment):
  authenticate()
  post_url = f"{BASE_URL}/{customerpayment_id}"
  result = put(post_url, HEADERS, json.dumps(customerpayment))
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = put(post_url, HEADERS, json.dumps(customerpayment))
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to update customerpayment. Result: {result}")
  return result


def get_customerpayment(customerpayment_id):
  authenticate()
  post_url = f"{BASE_URL}/{customerpayment_id}"
  result = get(post_url, HEADERS)
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to retrieve customerpayment. Result: {result}")
  return result


def search_customerpayments(customer_name=None,
                            reference_number=None,
                            date=None,
                            amount=None,
                            notes=None,
                            payment_mode=None,
                            due_date=None,
                            status=None,
                            customer_id=None,
                            filter_by=None,
                            search_text=None,
                            sort_column='date',
                            **kwargs):
  authenticate()
  # filter_by Allowed Values: Status.All, Status.Sent, Status.Draft, Status.OverDue, Status.Paid, Status.Void, Status.Unpaid, Status.PartiallyPaid, Status.Viewed and Date.PaymentExpectedDate
  # sort_column Allowed Values: customer_name, customerpayment_number, date, due_date, total, balance and created_time
  post_url = f"{BASE_URL}?sort_column={sort_column}"
  if customer_name is not None:
    post_url = f"{post_url}&customer_name={customer_name}"
  if reference_number is not None:
    post_url = f"{post_url}&reference_number={reference_number}"
  if date is not None:
    post_url = f"{post_url}&date={date}"
  if amount is not None:
    post_url = f"{post_url}&amount={amount}"
  if notes is not None:
    post_url = f"{post_url}&notes={notes}"
  if payment_mode is not None:
    post_url = f"{post_url}&payment_mode={payment_mode}"
  if due_date is not None:
    post_url = f"{post_url}&due_date={due_date}"
  if status is not None:
    post_url = f"{post_url}&status={status}"
  if customer_id is not None:
    post_url = f"{post_url}&customer_id={customer_id}"
  if filter_by is not None:
    post_url = f"{post_url}&filter_by={filter_by}"
  if search_text is not None:
    post_url = f"{post_url}&search_text={search_text}"
  for name, value in kwargs.items():
    post_url = f"{post_url}&{name}={value}"
  result = get(post_url, HEADERS)
  if Utils.is_unauthorized(result['status_code']):
    authenticate(force=True)
    result = get(post_url, HEADERS)
  if not Utils.is_success(result['status_code']):
    raise Exception(f"Failed to retrieve customerpayments. Result: {result}")
  return result



from __future__ import print_function

import datetime


def create_task(project, location, queue, relative_handler_uri, payload=None, in_seconds=None):
  # [START cloud_tasks_appengine_create_task]
  """Create a task for a given queue with an arbitrary payload."""

  from google.cloud import tasks_v2
  from google.protobuf import timestamp_pb2

  # Create a client.
  client = tasks_v2.CloudTasksClient()

  # Construct the fully qualified queue name.
  parent = client.queue_path(project, location, queue)

  # Construct the request body.
  task = {
      'app_engine_http_request': {  # Specify the type of request.
          'http_method': 'POST',
          'relative_uri': relative_handler_uri
      }
  }
  if payload is not None:
    # The API expects a payload of type bytes.
    converted_payload = payload.encode()

    # Add the payload to the request.
    task['app_engine_http_request']['body'] = converted_payload

  if in_seconds is not None:
    # Convert "seconds from now" into an rfc3339 datetime string.
    d = datetime.datetime.utcnow() + datetime.timedelta(seconds=in_seconds)

    # Create Timestamp protobuf.
    timestamp = timestamp_pb2.Timestamp()
    timestamp.FromDatetime(d)

    # Add the timestamp to the tasks.
    task['schedule_time'] = timestamp

  # Use the client to build and send the task.
  response = client.create_task(parent, task)
  return response
# [END cloud_tasks_appengine_create_task]
